/**
 * This file is part of GenomeView, a genome browser and annotation curator
 * 
 * Copyright (C) 2007-2013 Thomas Abeel and contributors
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 * Project: http://genomeview.org/
 * 
 * Copyright 2004, Generation5. All Rights Reserved.
 */
package net.sf.genomeview.core;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

/**
 * This class provides a method of creating a smooth gradient evenly distributed
 * colour points. To create a gradient, add as many colors as you like and call
 * <code>createGradient</code>. <code>getColour</code> or
 * <code>getGradient</code> can be used to retrieve the resultant colors.
 * 
 * @author James Matthews
 * @author Thomas Abeel
 */
public class ColorGradient {
	/**
	 * Default 512 step four color gradient (red, yellow, green, blue)
	 */
	public static ColorGradient fourColorGradient = null;

	/**
	 * Default 512 step two color gradient (red, green)
	 */
	public static ColorGradient redGreenColorGradient = null;

	/**
	 * Default 512 step four color gradient (yellow, blue);
	 */
	public static ColorGradient yellowBlueGradient = null;

	static {
		fourColorGradient = new ColorGradient();
		fourColorGradient.addPoint(Color.red);
		fourColorGradient.addPoint(Color.yellow);
		fourColorGradient.addPoint(Color.green);
		fourColorGradient.addPoint(Color.blue);
		fourColorGradient.createGradient(512);

		redGreenColorGradient = new ColorGradient();
		redGreenColorGradient.addPoint(Color.green);
		redGreenColorGradient.addPoint(Color.red);
		redGreenColorGradient.createGradient(512);

		yellowBlueGradient = new ColorGradient();
		yellowBlueGradient.addPoint(Color.yellow);
		yellowBlueGradient.addPoint(Color.blue);
		yellowBlueGradient.createGradient(512);

	}

	private List<Color> crGradientPoints = new ArrayList<Color>();
	private List<Color> crGradient;

	/**
	 * Add a color to the gradient list.
	 * 
	 * @param gradientColour
	 *            the next color in the gradient.
	 */
	public void addPoint(Color gradientColour) {
		crGradientPoints.add(gradientColour);
	}

	/**
	 * Creates a 512 step 2 color gradient with the supplied start and end
	 * colors.
	 * 
	 * @param start
	 *            starting color
	 * @param end
	 *            ending color
	 * @return a gradient between the two colors
	 */
	public static ColorGradient createTwoColor(Color start, Color end) {
		ColorGradient out = new ColorGradient();
		out.addPoint(start);
		out.addPoint(end);
		out.createGradient(512);
		return out;
	}

	/**
	 * Calculate the RGB deltas between two different color values and over a
	 * given number of steps. <code>createGradient</code> uses this function to
	 * connect the gradient points together.
	 * 
	 * @param start
	 *            the starting color.
	 * @param end
	 *            the ending color.
	 * @param steps
	 *            the number of steps required to get from <code>start</code> to
	 *            <code>end</code.
	 * @return a <code>double[3]</code> array returning the red, green and blue
	 *         delta values.
	 */
	private double[] getRGBDeltas(Color start, Color end, int steps) {
		double[] delta = new double[3];

		delta[0] = (end.getRed() - start.getRed()) / (double) steps;
		delta[1] = (end.getGreen() - start.getGreen()) / (double) steps;
		delta[2] = (end.getBlue() - start.getBlue()) / (double) steps;

		return delta;
	}

	/**
	 * Create the gradient using the current gradient list. This method defaults
	 * to using 256 steps.
	 */
	public void createGradient() {
		createGradient(256);
	}

	/**
	 * Create the gradient using the current gradient list.
	 * 
	 * @param numSteps
	 *            the total number of steps to take from the first color to the
	 *            last.
	 */
	public void createGradient(int numSteps) {
		int steps = numSteps / (crGradientPoints.size() - 1);
		double[] crColours = new double[3];
		crColours[0] = crGradientPoints.get(0).getRed();
		crColours[1] = crGradientPoints.get(0).getGreen();
		crColours[2] = crGradientPoints.get(0).getBlue();

		crGradient = new ArrayList<Color>();// new Color[numSteps];

		// For each of the gradient points
		for (int i = 0; i < crGradientPoints.size() - 1; i++) {
			double[] delta = getRGBDeltas(crGradientPoints.get(i), crGradientPoints.get(i + 1), steps);

			for (int s = 0; s < steps; s++) {
				crColours[0] += delta[0];
				crColours[1] += delta[1];
				crColours[2] += delta[2];
				if (crColours[0] > 255)
					crColours[0] = 255;
				if (crColours[1] > 255)
					crColours[1] = 255;
				if (crColours[2] > 255)
					crColours[2] = 255;
				crGradient.add(new Color((int) Math.round(crColours[0]), (int) Math.round(crColours[1]), (int) Math.round(crColours[2])));

			}
		}
	}

	/**
	 * Return the <i>ith</i> colour in the gradient.
	 * 
	 * @param i
	 *            the index of gradient array to return.
	 * @return the <code>Color</code> value in the gradient.
	 */
	public Color getColor(int i) {
		if(i<0)
			i=0;
		if(i>=crGradient.size())
			i=crGradient.size()-1;
		return crGradient.get(i);
	}
	public Color getColor(double s) {
		return getColor((int)(s*crGradient.size()));
	}

	/**
	 * Return the entire gradient as a Color list.
	 * 
	 * @return list containing the gradient colors.
	 */
	public List<Color> getGradient() {
		return crGradient;
	}

	/**
	 * Test function.
	 * 
	 * @param args
	 *            command-line arguments (ignored).
	 */
	public static void main(String args[]) {
		ColorGradient gradient = new ColorGradient();

		gradient.addPoint(new Color(64, 0, 128));
		gradient.addPoint(new Color(255, 0, 128));
		gradient.addPoint(new Color(255, 255, 128));
		gradient.addPoint(Color.WHITE);
		gradient.createGradient(512);

		BufferedImage buffer = new BufferedImage(512, 100, 1);
		Graphics graphics = buffer.createGraphics();

		for (int i = 0; i < 512; i += 2) {
			graphics.setColor(gradient.getColor(i / 2));
			graphics.fillRect(i, 0, 2, 100);
		}

		java.awt.image.RenderedImage rendered = buffer;
		try {
			File file = new File("gradtest.png");
			ImageIO.write(rendered, "png", file);
		} catch (IOException e) {
			System.err.println("An error occurred.");
		}
	}

	
}
